-------------------------------------------------------------------------------
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111-1307 USA
-------------------------------------------------------------------------------

include get.e
include misc.e
include std/console.e as con
include std/graphics.e
include std/sequence.e
include std/types.e

constant colors = {"blue", "green", "teal", "red", "pink", "brown", "white", "gray",
                   "bright blue", "bright green", "bright teal", "bright red", "bright pink", "yellow", "bright white"},
         names = {"Vinny", "Blake", "Gina", "Matt", "Sigmund", "Wanda",
                  "Pazuzu", "Lilith", "Icarus", "Hera", "Thoth", "G.G.",
                  "Raymond", "Xerxes", "Ned", "Kurt", "Lulu", "Hannah",
                  "Xavier", "Mik", "Antoine", "Andre", "Eric", "O'Shea",
                  "Lorenzo", "Janis", "Daria", "Simone", "Ursula", "Egbert",
                  "Quinn", "Jake", "Helen", "Greg"},

         M_SNIFF        = 1,
         M_HEALTH       = 2,
         M_LIFE         = 3,
         M_MAX_LIFE     = 4,
         M_TRIED        = 5,
         M_STACK        = 6,
         M_LAST         = 7,
         M_POSITION     = 8,
         M_GENERATION   = 9,
         M_NAME         = 10,
         M_COLOR        = 11,
         M_SNIFFING     = 12,

         PLATFORM       = platform()

procedure delay (atom d)
    atom t
    t = time()

    while time () < t + d do end while
end procedure

-- format of entries in stack:
--  
--   last position,
--   choices at that position,
--   tried cells at that position
--  }

sequence creatures, path, food, maze, temp, remove_list, options, empty, MAZE_FILE
integer a, k, last_color, quit, pass_through, NUM_CREATURES, BASE_LIFE
atom next_feed, next_sub, wait, exec, mut_rate,
     FEED_INTERVAL, SUB_INTERVAL, FOOD_DURATION, TURN_DELAY

MAZE_FILE = "maze.txt"
NUM_CREATURES        = 1
FEED_INTERVAL   = 2.5
SUB_INTERVAL    = 1
FOOD_DURATION   = 1
TURN_DELAY      = 0.05
BASE_LIFE       = 1000

mut_rate = 7
pass_through = 0
quit = 1
last_color = 1
creatures = {}
food = {}
remove_list = {}
path = {}
empty = {}

-- pick a name out of the hat
function pick_name ()
    integer n, o

    while 1 do
        -- select a name at random
        n = rand(length(names))
        o = 0

        -- is it taken?
        for i = 1 to length(creatures) do
            if equal(creatures[i][M_NAME], names[n]) then
                o = 1
            end if
        end for

        if not o then
            -- if our name isn't used, we can stop picking
            exit
        end if

    -- otherwise, go back and try again
    end while

    return names[n]
end function

-- pick a random position for a new creature
function pick_new ()
    integer m, n
    sequence pos

    while 1 do
        -- select one of the empty cells at random
        n = rand(length(empty))
        pos = empty[n]
        m = 0

        -- iterate through the living creatures
        for i = 1 to length(creatures) do
            -- is this one living at the spot we picked?
            if equal(creatures[i][M_POSITION], pos) then
                m = 1
                exit
            end if

        -- go on to the next one
        end for

        if m = 0 then
            -- if not, return coordinates of the cell we picked
            return empty[n]
        end if
    end while
end function

function pick_color ()
    integer c, count, x

    c = last_color

    while 1 do
        x = 0
        count = 0
        c += 1
        if c = 16 then
            c = BLUE
        elsif c = WHITE then
            c = GRAY
        end if

        for i = 1 to length(creatures) do
            if creatures[i][M_COLOR] = c then
                x = 1
                count += 1
                exit
            end if
        end for
        
        if not x or count = length(creatures) then
            exit
        end if
    end while

    last_color = c
    return c
end function

-- create a new creature
procedure new_mouse (sequence position = {}, integer color = -1,
                     integer sniff_distance = 3, integer life = -1, integer generation = 1)
    if length(position) = 0 then
        position = pick_new()
    end if

    if color = -1 then
        color = pick_color()
    end if

    if life = -1 then
        life = BASE_LIFE
    end if

    -- set sniff_distance to 1 if an creature tries to spawn offspring with no
    -- sniff distance
    if sniff_distance < 1 then
        sniff_distance = 1
    end if

    -- add creature to the list with default attributes and a unique name and
    -- position in the maze
    creatures = append(creatures, {sniff_distance, 100, life, life, {}, {}, {}, position,
                         generation, pick_name(), color, 0})
    creatures[$][M_LAST] = creatures[$][M_POSITION]
end procedure

-- read in the maze from a file 
procedure setup ()
    integer fn
    object ln
    sequence l, r, u, d

    fn = open(MAZE_FILE, "r")
    maze = {}

    while 1 do
        ln = gets(fn)
        if atom(ln) then exit end if
        if find('\n', ln) then ln = ln[1..$ - 1] end if

        maze &= {ln}
    end while
    close(fn)

    path = repeat(repeat({}, length(maze[1])), length(maze))

    for i = 1 to length(maze) do
        for j = 1 to length(maze[i]) do
            if maze[i][j] = ' ' then
                empty &= {{i, j}}
            end if
        end for
    end for

    for i = 1 to length(maze) do
        for j = 1 to length(maze[i]) do
            if i > 1 and maze[i - 1][j] = ' ' then
                path[i][j] &= {{i - 1, j}}
            end if
            
            if i < length(maze) and maze[i + 1][j] = ' ' then
                path[i][j] &= {{i + 1, j}}
            end if

            if j > 1 and maze[i][j - 1] = ' ' then
                path[i][j] &= {{i, j - 1}}
            end if

            if j < length(maze[1]) and maze[i][j + 1] = ' ' then
                path[i][j] &= {{i, j + 1}}
            end if
        end for
    end for

    cursor(NO_CURSOR)
end procedure

-- check if a cell is occupied by another creature
function occupied (integer r, integer c)
    if pass_through then
        return 0
    end if

    for i = 1 to length(creatures) do
        if equal(creatures[i][M_POSITION], {r, c}) then
            return 1
        end if
    end for

    return 0
end function

-- check if a cell has food in it
function has_food (integer r, integer c)
    for i = 1 to length(food) do
        if sequence(food[i]) and equal(food[i][1], {r, c}) then
            return i
        end if
    end for

    return 0
end function

-- sniff the surrounding area for food
-- sniff_food() traverses the area recursively, emitting "feelers" that work
-- their way through the maze, looking for food
function sniff_food (integer r, integer c, integer level,
                     sequence backtrack = {})
    sequence found_food, choices, s

    -- if we've used up our sniff ability, stop searching in this direction
    if level < 1 then
        return {}
    end if

    -- does this cell have food?
    if has_food(r, c) then
        position(r, c)
        text_color(BRIGHT_WHITE) bk_color(RED)
        puts(1, '&')
        text_color(WHITE) bk_color(BLACK)
        -- stop searching this path and report back
        return {{{r, c}, backtrack, level}}
    end if

    found_food = {}
    -- get path cells in surrounding area
    choices = path[r][c]

    -- search each one
    for i = 1 to length(choices) do
        -- check that we haven't been to this cell
        if not find(choices[i], backtrack)
        -- check that it's not occupied
        and not occupied(choices[i][1], choices[i][2]) then
            -- sniff for food at this cell
            s = sniff_food(choices[i][1], choices[i][2], level - 1,
                           {{r, c}} & backtrack)

            -- add any food found to the list
            if length(s) then
                found_food &= s
            end if
        end if
    end for

    -- take just the second entry of the backtrack queue. the first entry is
    -- the position of the creature sniffing for food, and the rest of the entries
    for i = 1 to length(found_food) do
        if length(found_food[i][2]) >= 2 then
            found_food[i][1] = found_food[i][2][$ - 1]
        end if
    end for

    return found_food
end function

-- determine which food source is the closest
function closest_food (sequence lfood)
    integer max, index, level

    max = 0
    index = 0

    for i = 1 to length(lfood) do
        level = lfood[i][3]

        if level > max then
            max = level
            index = i
        end if
    end for

    return index
end function

procedure show_guide ()
    for i = 1 to length(maze) do
        position(i, length(maze[1]) + 1)
        puts(1, i + '@')
    end for

    position(length(maze) + 1, 1)
    for i = 1 to length(maze[1]) do
        puts(1, i + '@')
    end for
end procedure

-- print the maze to the screen
procedure show_maze ()
    position(1, 1)
    for i = 1 to length(maze) do
        for j = 1 to length(maze[1]) do
            if maze[i][j] = ' ' then
                bk_color(BLACK)
            elsif maze[i][j] = '*' then
                bk_color(GRAY)
            end if

            puts(1, ' ')
        end for
        puts(1, '\n')
    end for

    bk_color(BLACK)
    show_guide()
end procedure

function remove_occupied (sequence choices)
    sequence a

    for i = 1 to length(creatures) do
        a = creatures[i][M_POSITION]
        if find(a, choices) then
            choices = remove_item(a, choices)
        end if
    end for

    return choices
end function

procedure winner ()
    integer winner, generation, color

    generation = 0

    -- identify the winning species and the generation of its newest member
    for i = 1 to length(creatures) do
        if creatures[i][M_HEALTH] > 0 and creatures[i][M_GENERATION] > generation then
            generation = creatures[i][M_GENERATION]
            color = creatures[i][M_COLOR]
        end if
    end for

    position(length(maze) + 3, 1)
    bk_color(color)
    if color = GREEN or color = CYAN or color >= BRIGHT_BLUE then
        text_color(BLACK)
    else
        text_color(WHITE)
    end if

    puts(1, colors[color])

    bk_color(BLACK)
    text_color(WHITE)
    printf(1, " wins the simulation after %d generations. congratulations!\n", generation)

    abort(0)
end procedure

-- explore the maze
procedure explore (integer mouse)
    sequence tried, last, stack, choices, sniff, name, temp
    integer row, col, n, health, life, max_life, sniff_distance, generation,
            color, task_id, t, u

    -- retrieve creature attributes

    sniff_distance  = creatures[mouse][M_SNIFF]
    health          = creatures[mouse][M_HEALTH]
    life            = creatures[mouse][M_LIFE]
    max_life        = creatures[mouse][M_MAX_LIFE]
    tried           = creatures[mouse][M_TRIED]
    stack           = creatures[mouse][M_STACK]
    last            = creatures[mouse][M_LAST]
    row             = creatures[mouse][M_POSITION][1]
    col             = creatures[mouse][M_POSITION][2]
    generation      = creatures[mouse][M_GENERATION]
    name            = creatures[mouse][M_NAME]
    color           = creatures[mouse][M_COLOR]

    if life <= 0 or health <= 0 then
        -- if this creature is dead, mark it for removal
        remove_list &= mouse

        -- and turn it into food!
        food &= {{{row, col}, time() + FOOD_DURATION}}

        -- last lineage left?
        t = color
        u = 0

        for i = 1 to length(creatures) do
            if creatures[i][M_COLOR] != t then
                u = 1
                exit
            end if
        end for

        if u = 0 and quit then
            winner()
        end if

        return
    end if

    -- add current cell to list of tried cells
    tried &= {{row, col}}

    -- forget all but the last 7
    if length(tried) >= 7 then tried = tried[$ - 6..$] end if

    -- scan for empty cells around current position
    choices = path[row][col]

    -- exclude any cells already tried
    for i = 1 to length(tried) do
        choices = remove_item(tried[i], choices)
    end for

    choices = remove_occupied(choices)

    -- check if this cell has food
    n = has_food(row, col)
    if n then
        -- eat it!
        health += 15
        food = remove_item(food[n], food)
    end if

    -- if we have enough health to spawn...
    if health >= 200 then
        temp = path[row][col]

        if length(temp) then
            -- give half the health to a new creature
            health = floor(health / 2)
            life = floor(life * .67)

            -- pick a different color a certain percentage of the time
            if rand(100) <= mut_rate then
                new_mouse(temp[rand(length(temp))],
                          -1,
                          -- this one might be better or worse at sniffing than
                          -- its parent, by up to +/-2
                          sniff_distance + (rand(5) - 3),
                          floor(max_life * (rand(75) + 45) / 100),
                          generation + 1)
            else
                new_mouse(temp[rand(length(temp))],
                          color,
                          sniff_distance + (rand(5) - 3),
                          floor(max_life * (rand(75) + 45) / 100),
                          generation + 1)
            end if
        end if
    end if

    -- sniff for food in the vicinity
    sniff = sniff_food(row, col, sniff_distance + 1)

    -- found food?
    if length(sniff) then
        -- get index of closest food source
        n = closest_food(sniff)

        -- is it occupied?
        if not occupied(sniff[n][1][1], sniff[n][1][2]) then
            row = sniff[n][1][1]
            col = sniff[n][1][2]

            -- push the current position and remaining choices onto the stack
            stack &= {{{row, col}, remove_item({row, col}, choices)}}
        end if
    -- no, see where we can wander that we haven't been yet
    elsif length(choices) then
        -- pick one at random
        temp = choices[rand(length(choices))]
        -- push the current position and remaining choices onto the stack
        stack &= {{{row, col}, remove_item(temp, choices)}}

        -- move to the position we chose
        row = temp[1]
        col = temp[2]
    -- no empty cells
    elsif length(stack) then
        -- we have to backtrack

        -- first, look at the top of the stack. it contains the last location
        -- of the mouse, but that position could now be occupied. is it?
        if not occupied(stack[$][1][1], stack[$][1][2]) then
            -- no - take choices and position from top of stack
            row = stack[$][1][1]
            col = stack[$][1][2]
            choices = stack[$][2]
            -- and pop
            stack = stack[1..$ - 1]

            -- were there any choices at this position?
            if length(choices) then
                -- pick one at random
                n = rand(length(choices))

                row = choices[n][1]
                col = choices[n][2]
                choices = remove_item({row, col}, choices)
            end if
        else
            -- yes - erase tried cells, allowing it to at least move to an path cell
            tried = {}            
        end if
    -- no empty cells, and stack is exhausted
    else
        -- start over
        tried = {}
    end if

    if length(last) then
        -- erase mouse at old position
        position(last[1], last[2])
        puts(1, ' ')

        -- draw at new position
        position(row, col)
        bk_color(color)
        if find(color, {CYAN, GREEN}) or color >= BRIGHT_BLUE then
            text_color(BLACK)
        else
            text_color(BRIGHT_WHITE)
        end if
        puts(1, name[1])
        text_color(WHITE)
        bk_color(BLACK)
    end if

    -- save creature attributes
    creatures[mouse] = {sniff_distance, health, life, max_life, tried, stack,
                   {row, col}, {row, col}, generation, name, color, length(sniff)}
end procedure

-- deposit sustenance into the maze
procedure deposit_food ()
    integer n

    while 1 do
        -- pick a random empty cell
        n = rand(length(empty))

        -- check that it's not occupied by an creature
        if not occupied(empty[n][1], empty[n][2]) then
            -- draw food there
            position(empty[n][1], empty[n][2])
            text_color(BLACK) bk_color(WHITE)
            puts(1, '&')
            bk_color(BLACK) text_color(WHITE)

            -- add to the list
            food &= {{empty[n], time() + FOOD_DURATION}}

            return
        end if
    end while
end procedure

function color_code (integer i)
    if i < 25 then
        return RED
    elsif i < 50 then
        return BROWN
    elsif i < 75 then
        return GREEN
    elsif i < 100 then
        return BRIGHT_GREEN
    elsif i < 145 then
        return CYAN
    elsif i < 185 then
        return BRIGHT_CYAN
    end if

    return BRIGHT_WHITE
end function

procedure creatures_stats ()
    sequence life, sniff, pop
    integer x, top_life, top_sniff, top_population, ts_color, tl_color, tp_color

    life = {}
    sniff = {}

    for i = 1 to length(creatures) do
        x = 0

        for j = 1 to length(life) do
            if life[j][1] = creatures[i][M_COLOR] then
                life[j][2] += creatures[i][M_MAX_LIFE]
                life[j][3] += 1
                x = 1
                exit
            end if
        end for

        if not x then
            life &= {{creatures[i][M_COLOR], creatures[i][M_MAX_LIFE], 1}}
        end if
    end for

    for i = 1 to length(creatures) do
        x = 0

        for j = 1 to length(sniff) do
            if sniff[j][1] = creatures[i][M_COLOR] then
                sniff[j][2] += creatures[i][M_SNIFF]
                sniff[j][3] += 1
                x = 1
                exit
            end if
        end for

        if not x then
            sniff &= {{creatures[i][M_COLOR], creatures[i][M_SNIFF], 1}}
        end if
    end for

    top_life = 0
    for i = 1 to length(life) do
        life[i][2] /= life[i][3]
        if life[i][2] > top_life then
            top_life = floor(life[i][2])
            tl_color = life[i][1]
        end if
    end for

    top_sniff = 0
    for i = 1 to length(sniff) do
        sniff[i][2] /= sniff[i][3]
        if sniff[i][2] > top_sniff then
            top_sniff = floor(sniff[i][2])
            ts_color = sniff[i][1]
        end if
    end for

    pop = {}
    for i = 1 to 15 do
        pop &= 0
        for j = 1 to length(creatures) do
            if creatures[j][M_COLOR] = i then
                pop[$] += 1
            end if
        end for
    end for

    top_population = 0
    for i = 1 to length(pop) do
        if pop[i] > top_population then
            top_population = pop[i]
            tp_color = i
        end if
    end for

    bk_color(tl_color)
    if find(tl_color, {GREEN, CYAN}) or tl_color >= BRIGHT_BLUE then
        text_color(BLACK)
    else
        text_color(BRIGHT_WHITE)
    end if
    position(length(creatures) + 3, length(maze[1]) + 3)
    puts(1, colors[tl_color])

    text_color(WHITE)
    bk_color(BLACK)
    printf(1, " is average longest lived (%d)                    ", top_life)

    bk_color(ts_color)
    if find(ts_color, {GREEN, CYAN}) or ts_color >= BRIGHT_BLUE then
        text_color(BLACK)
    else
        text_color(BRIGHT_WHITE)
    end if
    position(length(creatures) + 4, length(maze[1]) + 3)
    puts(1, colors[ts_color])

    text_color(WHITE)
    bk_color(BLACK)
    printf(1, " has the average farthest sniffers (%g)                    ", top_sniff)

    position(length(creatures) + 5, length(maze[1]) + 3)
    bk_color(tp_color)
    if find(tp_color, {GREEN, CYAN}) or tp_color >= BRIGHT_BLUE then
        text_color(BLACK)
    else
        text_color(BRIGHT_WHITE)
    end if
    puts(1, colors[tp_color])

    text_color(WHITE)
    bk_color(BLACK)
    puts(1, " is most numerous                   ")

    position(length(creatures) + 6, length(maze[1]) + 3)
    puts(1, repeat(32, 80))
end procedure

-- display vital statistics on living creatures next to the maze
-- stats given: name, generation, health level, sniffing distance
procedure creatures_info ()
    integer left, color

    left = length(maze[1]) + 3

    position(1, length(maze[1]) + 3)
    puts(1, "#  NAME     GEN  HLTH  LIFE  SNIFF")

    for i = 1 to length(creatures) do
        color = creatures[i][M_COLOR]
        position(i + 1, left)

        text_color(BLACK)
        bk_color(color)
        printf(1, "%d", i)

        bk_color(BLACK)
        puts(1, ' ')

        bk_color(color)
        position(i + 1, left + 3)
        printf(1, "%s", {creatures[i][M_NAME]})

        bk_color(BLACK)
        text_color(WHITE)
        puts(1, repeat(' ', 9 - length(creatures[i][M_NAME])))

        position(i + 1, left + 12)
        printf(1, "%d ", creatures[i][M_GENERATION])

        position(i + 1, left + 17)
        bk_color(color_code(creatures[i][M_HEALTH]))
        text_color(BLACK)
        printf(1, "%d", creatures[i][M_HEALTH])
        bk_color(BLACK)
        text_color(WHITE)
        puts(1, "  ")

        position(i + 1, left + 23)
        bk_color(color_code(creatures[i][M_LIFE]))
        text_color(BLACK)
        printf(1, "%d", creatures[i][M_LIFE])
        bk_color(BLACK)
        text_color(WHITE)
        puts(1, "  ")

        position(i + 1, left + 29)
        printf(1, "%d", creatures[i][M_SNIFF])

        if creatures[i][M_SNIFFING] then
            puts(1, ' ' & repeat('*', length(creatures[i][M_SNIFFING])))
        end if

        puts(1, "     ")
    end for

    bk_color(BLACK)
    text_color(WHITE)
    position(length(creatures) + 2, length(maze[1]) + 3)
    puts(1, repeat(32, 80))

    creatures_stats()
end procedure

-- save the current maze and its inhabitants to disk
procedure save_state ()
    sequence fname
    integer fn, bz2

    bz2 = 0
    position(length(maze) + 2, 1)

    fname = con:prompt_string("file name: ")
    if length(fname) = 0 then
        return
    end if

    if match(".bz2", fname) = length(fname) - 3 then
        fname = fname[1..$ - 4]
        bz2 = 1
    end if

    fn = open(fname, "w")
    if fn = -1 then
        return
    end if

    print(fn, {maze, creatures, food})
    close(fn)

    if bz2 then
        -- delete the original compressed save state
        if PLATFORM = LINUX or PLATFORM = FREEBSD then
            system("rm " & fname & ".bz2", 2)
        else
            system("del " & fname & ".bz2", 2)
        end if

        -- compress the new one
        system("bzip2 " & fname, 2)
    end if

    position(length(maze) + 2, 1)
    puts(1, repeat(32, 80))
end procedure

-- load a maze and inhabitants from disk
procedure load_state ()
    sequence fname, state
    integer fn, bz2

    bz2 = 0
    position(length(maze) + 2, 1)

    fname = con:prompt_string("file name: ")
    if length(fname) = 0 then
        return
    end if

    -- decompress first if we're given a bzipped save state
    if match(".bz2", fname) = length(fname) - 3 then
        system("bunzip2 " & fname, 2)
        fname = fname[1..$ - 4]

        -- and tie a string around the index finger
        bz2 = 1
    end if

    fn = open(fname, "r")
    if fn = -1 then
        return
    end if

    state = get(fn)
    close(fn)

    -- do we have a string on our index finger?
    if bz2 then
        -- recompress the save state we just loaded
        system("bzip2 " & fname, 2)
    end if

    if state[1] = GET_SUCCESS then
        state = state[2]

        maze = state[1]
        creatures = state[2]
        food = state[3]

        clear_screen()
        show_maze()
    end if

    position(length(maze) + 2, 1)
    puts(1, repeat(32, 80))
end procedure

procedure help ()
    sequence interpreter, program, command

    command = command_line()
    interpreter = command[1]
    program = command[2]

    if equal(interpreter, command) then
        printf(1, "usage: %s ", {program})
    else
        printf(1, "usage: eui %s ", {program})
    end if

    printf(1, "[-h] [-c] [-l N] [-m F] [-n N] [-d N] [-f N] [-s N]\n\n" &
              "  -h  \tdisplay this help page\n" &
              "  -c  \tcontinue the simulation even after one lineage achieves domination\n" &
              "  -m F\tload maze file F (default: %s)\n" &
              "  -l N\tset the base life value (default: %d)\n" &
              "  -n N\tstart with N creatures (default: %d)\n" &
              "  -t N\tturn delay in seconds (default: %g)\n"&
              "  -d N\tduration of food (default: %g)\n" &
              "  -f N\tfeeding interval (default: %g)\n" &
              "  -s N\thealth decrement interval (default: %g)\n",
           {MAZE_FILE, BASE_LIFE, NUM_CREATURES, TURN_DELAY, FOOD_DURATION,
            FEED_INTERVAL, SUB_INTERVAL})
end procedure

options = command_line()

for i = 3 to length(options) do
    if find(options[i], {"-h", "--help"}) then
        help()
        abort(0)
    elsif equal(options[i], "-c") then                          -- continue mode
        quit = 0
    elsif equal(options[i], "-p") then                          -- passthrough mode
        pass_through = 1
    elsif equal(options[i], "-m") and i < length(options) then  -- maze file
        MAZE_FILE = options[i + 1]

    -- these arguments may take floating point values
    elsif equal(options[i], "-t") and i < length(options) then  -- turn delay
        options[i + 1] = value(options[i + 1])
        if options[i + 1][1] = GET_SUCCESS then
            TURN_DELAY = options[i + 1][2]
        end if
    elsif equal(options[i], "-d") and i < length(options) then  -- food duration
        options[i + 1] = value(options[i + 1])
        if options[i + 1][1] = GET_SUCCESS then
            FOOD_DURATION = options[i + 1][2]
        end if
    elsif equal(options[i], "-f") and i < length(options) then  -- feeding interval
        options[i + 1] = value(options[i + 1])
        if options[i + 1][1] = GET_SUCCESS then
            FEED_INTERVAL = options[i + 1][2]
        end if
    elsif equal(options[i], "-s") and i < length(options) then  -- health subtraction interval
        options[i + 1] = value(options[i + 1])
        if options[i + 1][1] = GET_SUCCESS then
            SUB_INTERVAL = options[i + 1][2]
        end if

    -- these arguments all take integer values
    elsif equal(options[i], "-M") and i < length(options)       -- mutation rate
          and t_digit(options[i + 1]) then
        options[i + 1] = value(options[i + 1])
        mut_rate = floor(options[i + 1][2])
    elsif equal(options[i], "-l") and i < length(options)       -- base life value
          and t_digit(options[i + 1]) then
        options[i + 1] = value(options[i + 1])
        BASE_LIFE = floor(options[i + 1][2])
    elsif equal(options[i], "-n") and i < length(options)       -- initial number of creatures
          and t_digit(options[i + 1]) then
        options[i + 1] = value(options[i + 1])
        NUM_CREATURES = floor(options[i + 1][2])
    end if
end for

setup()
for i = 1 to NUM_CREATURES do
    new_mouse({}, -1, 3 + (rand(3) - 2), BASE_LIFE * (rand(50) + 75) / 100)
end for
clear_screen()
show_maze()

next_feed = time() + FEED_INTERVAL
next_sub = time () + SUB_INTERVAL

while length(creatures) do
    k = get_key()
    if k = 'q' then
        position(length(maze) + 3, 1)
        exit
    elsif k = 's' then
        save_state()
    elsif k = 'l' then
        load_state()
    elsif k = 'p' then
        position(length(maze) + 2, 1)

        wait = time()
        any_key()
        wait = time() - wait

        for i = 1 to length(food) do
            if sequence(food[i]) then
                food[i][2] += wait
            end if
        end for

        position(length(maze) + 2, 1)
        puts(1, repeat(32, 80))
    end if

    if time() >= next_feed then
        next_feed = time() + FEED_INTERVAL
        deposit_food()
    end if

    if time () >= next_sub then
        next_sub = time() + SUB_INTERVAL
        for i = 1 to length(creatures) do
            creatures[i][M_HEALTH..M_LIFE] -= 1
        end for
    end if

    creatures_info()

    exec = time()
    for i = 1 to length(creatures) do
        explore(i)
    end for
    exec = time() - exec

    for i = 1 to length(food) do
        if sequence(food[i]) and time() >= food[i][2] then
            position(food[i][1][1], food[i][1][2])
            puts(1, ' ')
            food[i] = 0
        end if
    end for

    food = remove_item(0, food)

    for i = 1 to length(remove_list) do
        a = remove_list[i]
        creatures = remove_item(creatures[a], creatures)
        for j = i + 1 to length(remove_list) do
            if remove_list[j] > a then
                remove_list[j] -= 1
            end if
        end for
    end for

    remove_list = {}

    if exec < TURN_DELAY then
        delay(TURN_DELAY - exec)
    end if
end while
